"""General utilities."""

from gettext import translation
from typing import Union

import pycountry

from . import DIR


def parse_lang(lang: str) -> Union[None, str]:
    """Attempt to convert `lang` to its alpha_2 version, if not already in that
    format.
    """
    normalised = lang.strip().lower()
    if pycountry.languages.get(alpha_2=normalised):
        return normalised

    try:
        language = pycountry.languages.lookup(normalised.title())
        return language.alpha_2
    except LookupError:
        return None


def translate(locale: str) -> None:
    """Install gettext translations for `locale`."""
    translation(
        "messages",
        localedir=DIR / "locales",
        languages=[locale],
    ).install()



# pylint: disable=undefined-variable
# pylint: disable=unused-wildcard-import
# ruff: noqa: F405
# mypy: disable-error-code="name-defined"

"""Main script/entry point for wordlemini."""

import random
import sys
from itertools import chain
from math import floor
from typing import List, Union

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Grid, Horizontal, Vertical
from textual.widget import Widget
from textual.widgets import Button, Input, Label, MarkdownViewer, Rule

from . import ASSETS, CORPUS
from ._config import Config
from ._keyboards import *  # noqa
from ._statistics import Statistics, _GuessDistributionTD, _StatisticsTD
from ._util import translate

DARK = Config.get("settings", "dark")
LANG = Config.get("settings", "lang")
translate(LANG)


class WordlePage(App):
    """Textual app to play a wordle game."""

    CSS_PATH = ASSETS / "wordlestyles.tcss"
    BINDINGS = [
        Binding("ctrl+n", "new_game_keybind", "Start new game"),
    ]
    GRID_WIDTH = 5
    GRID_HEIGHT = 6

    words: List[str] = []

    @classmethod
    def apply_words(cls) -> None:
        """Gather all the words in a particular `<lang-code>.txt` file and append
        them to `WordlePage.words`.
        """
        with open(CORPUS / f"{LANG}.txt", encoding="utf-8") as f:
            for word in f:
                formatted_word = word[:-1]
                if len(formatted_word) == WordlePage.GRID_WIDTH:
                    cls.words.append(formatted_word)

    @property
    def main_widgets(self) -> List[Widget]:
        """Return a list of all the main widgets."""
        return [
            *list(chain(*self.grid)),
            *list(chain(*self.keyboard)),
            self.input,
            self.submit,
            self.delete,
            self.new,
        ]

    def __init__(self) -> None:
        """Initialise widgets and pick a random word to begin the first game."""
        super().__init__()
        self.grid = [
            [Label("") for i in range(WordlePage.GRID_WIDTH)]
            for j in range(WordlePage.GRID_HEIGHT)
        ]
        self.dark = DARK
        if not DARK:
            for label in list(chain(*self.grid)):
                label.styles.border = ("round", "black")
        self.keyboard = [
            [Button(key, name="key") for key in row]
            for row in globals().get(f"KEYBOARD_{LANG.upper()}") or KEYBOARD_EN
        ]
        self.input = Input(
            placeholder=_("Enter your guess"),
            type="text",
            restrict=r"[a-zA-Z]{0,5}",
        )
        self.submit = Button(f"{_("Guess")} [↵]", name="submit")
        self.delete = Button(f"{_("Del")} [←]", name="delete")
        self.new = Button(f"{_("New")} [^n]", name="new")
        self.quit = Button(f"{_("Quit")} [^c]", name="quit", variant="error")

        WordlePage.apply_words()
        self.word = random.choice(WordlePage.words).casefold()
        self.guess = ""
        self.row = 0

        self.nonexisting_chars: set[str] = set()
        self.wrong_pos_chars: set[str] = set()
        self.correct_pos_chars: set[str] = set()

    def compose(self) -> ComposeResult:
        """Provide Textual with all the widgets for the DOM."""
        yield Horizontal(
            Grid(*list(chain(*self.grid)), classes="word-grid"),
            Rule(orientation="vertical"),
            Grid(
                Vertical(
                    *[
                        Horizontal(Grid(*row, classes="keyboard-row"))
                        for row in self.keyboard
                    ],
                    id="keyboard",
                ),
                Vertical(
                    Vertical(self.input, id="input-container"),
                    Horizontal(
                        self.submit,
                        Rule(),
                        self.delete,
                        Rule(),
                        self.new,
                        Rule(),
                        self.quit,
                        id="button-row",
                    ),
                ),
                id="right-side",
            ),
        )
        self.input.focus()

    def set_disabled_state_of_all_widgets(self, state: bool) -> None:
        """Enable or disable all main widgets."""
        for widget in self.main_widgets:
            widget.disabled = state

    def action_new_game_keybind(self) -> None:
        """Start a new game via a keybind."""
        self.new.action_press()

    def new_game(self) -> None:
        """Begin a new game - which involves picking a new word and resetting
        all colour-related hints in the keyboard as well as user input in the
        grid.
        """
        txt = _("Started new game!")
        if self.guess != self.word:
            txt += f" ({_("The word was")} {self.word.upper()})"
        self.notify(txt, timeout=2.0)
        for label in [*list(chain(*self.grid))]:
            label.update("")
            label.styles.border = (
                ("round", "white") if DARK else ("round", "black")  # type: ignore
            )
        for row in self.keyboard:
            for button in row:
                if button.name == "key":
                    button.label = str(button.label)
                    button.styles.opacity = "100%"

        self.input.value = ""
        self.nonexisting_chars = set()
        self.wrong_pos_chars = set()
        self.correct_pos_chars = set()
        self.word = random.choice(WordlePage.words).casefold()
        self.row = 0
        self.set_disabled_state_of_all_widgets(False)
        self.input.focus()

    def update_grid(self, guess: str) -> None:
        """Update the current row of the grid. This is a more advanced algorithm
        compared to `update_char_statuses` and ensures the proper detection of
        misplaced letters"""
        word_letter_count = {
            char: self.word.count(char) for char in set(self.word)
        }

        # first pass - check for correct positions and reduce counts
        for i, char in enumerate(guess):
            label = self.grid[self.row][i]
            if char == self.word[i]:
                label.update(f"[green]{char.upper()}[/green]")
                label.styles.border = ("round", "green")
                self.correct_pos_chars.add(char.casefold())
                word_letter_count[char] -= 1
            else:
                label.update(
                    f"{char.upper()}"
                )  # change if needed in next pass
                label.styles.border = ("round", "grey")

        # second pass - check for wrong positions using remaining counts
        for i, char in enumerate(guess):
            label = self.grid[self.row][i]
            if (
                char != self.word[i]
                and char in self.word
                and word_letter_count[char] > 0
            ):
                label.update(f"[yellow]{char.upper()}[/yellow]")
                label.styles.border = ("round", "yellow")
                self.wrong_pos_chars.add(char.casefold())
                word_letter_count[char] -= 1
            elif char != self.word[i] and char not in self.word:
                self.nonexisting_chars.add(char.casefold())

        self.row += 1

    def update_keyboard(self) -> None:
        """Update the virtual keyboard."""
        for row in self.keyboard:
            for button in row:
                label = str(button.label).casefold()
                if label in self.correct_pos_chars:
                    button.label = f"[green]{label.upper()}[/green]"
                elif label in self.wrong_pos_chars:
                    button.label = f"[yellow]{label.upper()}[/yellow]"
                elif label in self.nonexisting_chars:
                    button.styles.opacity = "35%"

    def update_char_statuses(self, guess: str) -> None:
        """Update the status of the characters in `guess`."""
        for i, char in enumerate(guess):
            if char == self.word[i]:
                self.correct_pos_chars.add(char)
            elif char in self.word:
                self.wrong_pos_chars.add(char)
            else:
                self.nonexisting_chars.add(char)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle a button being pressed."""
        name = event.button.name
        if name == "key":
            return self.action_submit_char(event)
        if name == "delete":
            return self.action_delete_char(event)
        if name == "submit":
            return self.on_input_submitted()
        if name == "new":
            return self.new_game()
        if name == "quit":
            sys.exit()
        return None

    def on_input_submitted(
        self,
        event: Union[Input.Submitted, None] = None,
    ) -> None:
        """Handle a user submitting their word."""
        guess = str(self.input.value).casefold()
        if len(guess) != WordlePage.GRID_WIDTH:
            return self.notify(
                _("Guess must be 5 characters long."),
                severity="error",
                timeout=2.0,
            )
        if guess not in WordlePage.words:
            return self.notify(
                _("Guess is not in word list."), severity="error", timeout=2.0
            )
        self.guess = guess
        self.update_char_statuses(guess)
        self.update_grid(guess)
        self.update_keyboard()
        self.input.value = ""
        self.input.focus()

        done = False
        if guess == self.word:
            self.notify(
                f"{_("You guessed the word in")} {self.row} {_("attempts")}!",
                timeout=3.5,
            )
            done = True
            Statistics.add_entry(True, self.row)
        elif self.row > WordlePage.GRID_WIDTH:
            self.notify(
                f"{_("The word was")} {self.word.upper()}. {_("Better luck next time")}!",
                timeout=3.5,
            )
            Statistics.add_entry(False)
            done = True
        if done:
            self.guess = self.word
            self.set_disabled_state_of_all_widgets(True)
            self.new.disabled = False

        return None

    def action_submit_char(self, event: Button.Pressed) -> None:
        """Enter a character from the virtual keyboard into `self.input`."""
        self.input.focus()
        if len(self.input.value) >= WordlePage.GRID_WIDTH:
            return
        self.input.value = (
            self.input.value + str(event.button.label).casefold()
        )
        self.input.action_cursor_right_word()

    def action_delete_char(self, event: Button.Pressed) -> None:
        """Delete a character from `self.input`."""
        self.input.focus()
        if len(self.input.value) <= 0:
            return
        self.input.value = self.input.value[:-1]
        self.input.action_cursor_left()


class StatisticsPage(App):
    """Textual app to view a user's wordlemini statistics."""

    CSS_PATH = ASSETS / "statisticstyles.tcss"

    @staticmethod
    def get_markdown(stats: _StatisticsTD) -> str:
        """Format a markdown string for the stats TUI using the statistic data
        in the `stats` param.
        """
        if not stats["games"]:
            winstr = "(NaN%)"
            lossstr = winstr
        else:
            winstr = f"({floor(stats['wins'] / stats['games'] * 100)}%)"
            lossstr = f"({floor(stats['losses'] / stats['games'] * 100)}%)"
        guess_dist = Statistics.get_square_bars(
            _GuessDistributionTD(**stats["guess_distribution"])  # type: ignore
        )
        return f"""
# {_("Your wordlemini stats")}

## {_("Guess distribution")}
{guess_dist}

## {_("Game stats")}
{_("Games played")}: {stats["games"]}

{_("Wins")}: {f"{stats["wins"]} {winstr}"}

{_("Losses")}: {f"{stats["losses"]} {lossstr}"}
        """

    def __init__(self) -> None:
        super().__init__()
        self.dark = DARK

    def compose(self) -> ComposeResult:
        stats = Config.read()["statistics"]
        yield Horizontal(MarkdownViewer(StatisticsPage.get_markdown(stats)))


def initialise_wordlepage() -> None:
    """Initialise the WordlePage TUI."""
    WordlePage().run()


def initialise_statisticspage() -> None:
    """Initialise the StatisticsPage TUI."""
    StatisticsPage().run()


if __name__ == "__main__":
    initialise_wordlepage()



"""Play wordle in the command-line."""

from pathlib import Path

from platformdirs import user_documents_dir

DIR = Path(__file__).resolve().parent
ASSETS = DIR / "assets"
CORPUS = ASSETS / "corpus"
DOCS = Path(user_documents_dir())
LANGUAGES = ["en", "es", "de", "fr", "ru", "pt", "it", "nl", "cs"]



"""Handling of user statistics such as guess distribution and games played."""

from typing import Dict, TypedDict

from ._config import Config


class _GuessDistributionTD(TypedDict):
    """Guess distribution format."""

    g1: int
    g2: int
    g3: int
    g4: int
    g5: int
    g6: int


class _StatisticsTD(TypedDict):
    """Format for the statistics section in the program config."""

    games: int
    wins: int
    losses: int
    guess_distribution: _GuessDistributionTD


class Statistics:
    """Handles the modification of statistics."""

    MAX_SQUARES = 37
    SQUARE = "\U000025a0"

    @staticmethod
    def add_entry(win: bool, guesses: int = 0) -> None:
        """Update the `statistics` section of the config."""
        cfg = Config.read()
        cfg["statistics"]["games"] += 1
        if win:
            cfg["statistics"]["wins"] += 1
            cfg["statistics"]["guess_distribution"][f"g{guesses}"] += 1
        else:
            cfg["statistics"]["losses"] += 1

        Config.write(cfg)

    @staticmethod
    def get_square_bars(
        # this is actually `_GuessDistributionTD` but I annotated it as `Dict`
        # to make mypy shut up
        distribution: Dict[str, int],
    ) -> str:
        """Convert the values (guesses) of a guess distribution dictionary to
        a bar-graph like string, similar to how NYT Wordle displays your guess
        distribution.
        """
        max_value = max(distribution.values())

        def calculate_squares(value: int) -> str:
            """Calculate how many squares should represent `value` based on
            `max_value`.
            """
            if not max_value:
                return ""
            return Statistics.SQUARE * int(
                (value / max_value) * Statistics.MAX_SQUARES
            )

        return "\n".join(
            [
                f"{i}  {calculate_squares(distribution.get(f'g{i}', 0))} ({distribution.get(f'g{i}', 0)})\n"  # noqa
                for i in range(1, 7)
            ]
        )



"""Handling of program configuration."""

from json import dump, load
from os import makedirs, path
from typing import Any

from . import DIR, DOCS


class Config:
    """Class for interacting with wordlemini's configuration JSON."""

    BASE_CFG = DIR / "template.data.json"
    DOC_TOPLEVEL = DOCS / "wordlemini"
    DOC_CFG = DOC_TOPLEVEL / "data.json"

    @property
    def fp(self):
        """Return the path to the config file."""
        makedirs(Config.DOC_TOPLEVEL, exist_ok=True)
        if not path.exists(Config.DOC_TOPLEVEL):
            return (
                Config.BASE_CFG
            )  # Return base config (documents not available)
        if not path.exists(Config.DOC_CFG):
            # Make config in system documents
            with open(Config.DOC_CFG, "x", encoding="utf-8") as dest, open(
                Config.BASE_CFG, encoding="utf-8"
            ) as src:
                dump(load(src), dest, indent=4)
        return Config.DOC_CFG

    @staticmethod
    def read() -> Any:
        """Return the data at `Config().fp`."""
        with open(Config().fp, encoding="utf-8") as f:
            return load(f)

    @staticmethod
    def write(data: Any) -> None:
        """Write `data` to `Config().fp`."""
        with open(Config().fp, "w", encoding="utf-8") as f:
            return dump(data, f, indent=4)

    @staticmethod
    def get(sec: str, opt: str) -> Any:
        """Get the value at `sec` and `opt` in the config."""
        return Config.read()[sec][opt]

    @staticmethod
    def set(sec: str, opt: str, val: Any) -> None:
        """Set config at `sec` and `opt` to `val` and write it to the file system."""
        cfg = Config.read()
        cfg[sec][opt] = val
        return Config.write(cfg)



"""Provide command-line argument parsing/handling functionality."""

from argparse import ArgumentParser, Namespace
from logging import INFO, Formatter, StreamHandler, getLogger

from . import LANGUAGES
from .__main__ import initialise_statisticspage, initialise_wordlepage
from ._config import Config
from ._util import parse_lang

USAGE_INFO = """
Type `wordlemini -h` to view an information page.

Simply type `wordlemini` to initialise the text interface.

Supported languages (configurable with `wordlemini config lang <lang-or-code>`)
- Czech, German, English, Spanish, French, Italian, Dutch, Portugese, Russian
"""

logger = getLogger()
logger.setLevel(INFO)
console = StreamHandler()
formatter = Formatter("%(levelname)s: %(message)s")
console.setFormatter(formatter)
logger.addHandler(console)


def parse() -> None:
    """Initialise the argument parser and add arguments."""
    parser = ArgumentParser(
        description="wordlemini CLI tool", prog="wordlemini", epilog=USAGE_INFO
    )

    subparsers = parser.add_subparsers(dest="command")

    # wordlemini {config}
    config_parser = subparsers.add_parser(
        "config",
        help="Configure settings. Type wordlemini config -h to learn more.",
    )
    config_subparsers = config_parser.add_subparsers(dest="config_command")

    # wordlemini {config} {lang} <lang-code>
    lang_parser = config_subparsers.add_parser(
        "lang", help="Set the language. e.x. `wordlemini config lang es`."
    )
    lang_parser.add_argument(
        "language", type=str, help="Language code or name"
    )

    # wordlemini {config} {dark} <on/off>
    dark_parser = config_subparsers.add_parser(
        "dark",
        help="Enable/disable dark mode. e.x. `wordlemini config dark off`.",
    )
    dark_parser.add_argument(
        "dark", type=str, help="Dark mode state - on or off"
    )

    # wordlemini {stats}
    subparsers.add_parser("stats", help="View your game statistics.")

    args = parser.parse_args()
    handle_args(args)


def handle_args(args: Namespace) -> None:
    """Handle the arguments the user has entered."""
    if args.command is None:
        return initialise_wordlepage()
    if args.command == "config":
        if args.config_command == "lang":
            lang = parse_lang(args.language)
            if lang is None or lang not in LANGUAGES:
                return logger.error(
                    "Language is either invalid or not supported by wordlemini."
                )
            Config.set("settings", "lang", lang)
            return logger.info("Set language to %s.", lang)
        if args.config_command == "dark":
            dark = args.dark.strip().casefold()
            if dark == "on":
                val = True
            elif dark == "off":
                val = False
            else:
                return logger.error("Value must either be on or off")
            Config.set("settings", "dark", val)
            return logger.info("Turned dark mode %s.", dark)
    if args.command == "stats":
        return initialise_statisticspage()

    return None


if __name__ == "__main__":
    parse()
